#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>
#ifndef F_CPU
#warning "F_CPU war noch nicht definiert, wird nun mit 16MHz definiert"
#define F_CPU 16000000UL     
#endif
#include <util/delay.h>         


/************************************************************************/
/*                                                                      */
/*                      Debouncing 8 Keys                               */
/*                      Sampling 4 Times                                */
/*                      With Repeat Function                            */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                      danni@specs.de                                  */
/*                                                                      */
/*                		ergnzt um eine Matrix/Multiplexansteuerung     */
/*                		auf 8*x Tasten/LED's     						*/
/*                      von Jens Khn                                   */
/*                      Spalten hier an Port A, Vorwiderstand ab 50Ohm */
/*                      Zeilen hier an Port C, Vorwiderstand ab 180Ohm  */
/*                      Kreuzungspunkte Taster                          */
/*                        parallel mit LED, Katode an PortC             */
/*                      Erweiterung einfach an Spalten,                 */
/*                        eher nicht in den Zeilen wegen LED-Helligkeit */
/*                        und Handhabbarkeit der in-/outbytes           */
/************************************************************************/

#include <main.h>          
#include <key_led.h>         


uint8_t led_mpex;						//die aktive LED-Zeile 
uint8_t led_blink_frequenz = 2;			//um unterschiedliche Zustnde zu signalisieren, in Hz
uint16_t led_blink_counter = 0;
uint8_t led_blink_status = 0;


void alle_blinken(bool on_off)
{
unsigned char i=0, zeile;

if (on_off)
  zeile = 0xff;
else 
  zeile = 0;

for (i=0;i<SPALTENANZAHL;i++)
  led_mode[i]=zeile;

return;
}

void init_key_led(void)
{


  KEY_DDR |=  ALL_KEYS;                	// als Ausgang fr die Zeilen
  KEY_PORT &= ~ALL_KEYS;                // und auf 0, spter multiplexen, dann auf 1
  SPALTE_DDR |=  ALLE_SPALTEN       ;

#if defined (__AVR_ATmega644P__)
  	TCCR0B = (1<<CS02)|(1<<CS00);			//prescaler auswhlen divide by 1024
#else
 	TCCR0 = (1<<CS02)|(1<<CS00);			// divide by 1024
#endif

#if defined (__AVR_ATmega644P__)
  TIMSK0 = 1<<TOIE0;				// enable timer interrupt
#else
  TIMSK |= 1<<TOIE0;					// enable timer interrupt
#endif
}

void einschwingen(void)  //scheinbar je nach Beschaltung notwendig nach Port E/A-Wechsel, ein nop reicht nicht
{
uint8_t i=F_CPU/800000; //5 bei 8MHz reicht nicht
while (i--)
 	asm volatile ("nop");
}



 
ISR( TIMER0_OVF_vect )                            // every 10ms

{
  static uint8_t ct0, ct1, rpt;
  uint8_t i;
  uint8_t key_mpex;		//die aktive Tasten-Spalte
  TCNT0 = (uint8_t)(int16_t)-((F_CPU/TEILER) / 1024 * 10e-3  + 0.5);  // nun 2,5ms wegen der LED-Frequenz
 
  if ((led_mpex % TEILER ) == 0 )  			//Tastenabfrage jede 4. Multiplexrunde, da kommen wir wieder auf 10ms
  {											//aktuelle Ports zwischenspeichern
  	//erst die Zeilen auf Eingang schalten, Pullups ein und die Spalten auf Ausgang

  	ZEILE_DDR &= ~ALLE_ZEILEN;                //zeileneingnge auf Eingang
	asm volatile ("nop");
 	asm volatile ("nop"); 	 
  	ZEILE_PORT |= ALLE_ZEILEN;                //und die Pullups an
 	asm volatile ("nop");
 	asm volatile ("nop"); 	 
	einschwingen();
	//jetzt alle Spalten durcharbeiten - bei vielen Spalten knnte es ein Zeitproblem werden?
	for (key_mpex=0;key_mpex < SPALTENANZAHL;key_mpex++)
		{
		SPALTE_DDR = (1<<key_mpex);  //damit die strme durch die LED's nicht quer flieen
 	    asm volatile ("nop");			 //alle anderen Spalten auf Eingang
 	    asm volatile ("nop"); 	 
		SPALTE_PORT = ~ALLE_SPALTEN;  //keine Pullups, TriState 
  		einschwingen();
		ct0 = 0;
		ct1 = 0;
		rpt = 0;
		i = 0;
   											   //bei mehr als 8 Spalten an Hardware anpassen
		i = key_state[key_mpex] ^ ~KEY_PIN;                       // key changed ?
  		ct0 = ~( ct0 & i );                             // reset or count ct0
  		ct1 = ct0 ^ (ct1 & i);                          // reset or count ct1
  		i &= ct0 & ct1;                                 // count until roll over ?
  		key_state[key_mpex] ^= i;                                 // then toggle debounced state
  		key_press[key_mpex] |= key_state[key_mpex] & i;                     // 0->1: key press detect
		}
   ZEILE_DDR |= ALLE_ZEILEN;                				//zeilen auf Ausgang
   asm volatile ("nop"); 	
  }
  
   //die Blinkfunktion abarbeiten
   if (++led_blink_counter > (400 / led_blink_frequenz))  //400 da aller Interrupt 2,5ms
     {
		led_blink_counter=0;
		led_blink_status ^= 0xff;
	 }
    
													//wenn mehr als 8 Spalten, dann hier auf das Multiplexen der Zeilen umstellen
  ZEILE_PORT = ~(led_s[led_mpex] & ~(led_mode[led_mpex]&led_blink_status));    //wenn blinkbit gesetzt ist
  													//die LED-Spalte drauflegen knnte noch maskiert werden negiert weil L-aktiv
  SPALTE_DDR = (1<<led_mpex);		//spalte auf Ausgang
	    asm volatile ("nop"); 	  
  SPALTE_PORT = 1<<led_mpex++; 						//und die entsprechende Spalte aktiv schalten
  if (led_mpex >= SPALTENANZAHL)					//runterzhlen wre wohl effektiver 
    led_mpex = 0;
 
}
 
///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed. Each pressed key is reported
// only once
// zustzlicher Par: Spalte
//
uint8_t get_key_press( uint8_t key_mask, uint8_t spalte )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_press[spalte];                          // read key(s)
  key_press[spalte] ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}
 
